'''
Date: 2023-03-12 19:43:45
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-12 20:12:30
FilePath: /outlier/src/outlier/biz/events.py
'''
from enum import Enum

class Ops(Enum):
    Add = 0
    Rmv = 1
    Rcv = 2



'''
Date: 2023-03-13 19:17:55
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-14 18:22:12
FilePath: /outlier/test/test_encryption.py
'''
import unittest
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent / "src" ))

from outlier.encryption.protocol import DES, RSA, Base64Encrption, Encryption


class TestEncryption(unittest.TestCase):
    
    test_contents = [
            "Hello world",
            "你好",
            "{'timestamp': 1678706203.87447, 'cmd': 'enterroom', 'param': 'Create and enter room 1'}",
            "⚔️"
            "(*^▽^*)",
            open(__file__, encoding='utf-8').read(),
        ]
    
    
    def test_encryption(self):
        for test_str in TestEncryption.test_contents:
            origin = test_str.encode()
            encrypted = Encryption.encrypt(origin)
            # print(encrypted)
            decrypted = Encryption.decrypt(encrypted)
            self.assertEqual(origin, decrypted)
            
    def test_base64Encrption(self):
        for test_str in TestEncryption.test_contents:
            origin = test_str.encode()
            encrypted = Base64Encrption.encrypt(origin)
            # print(encrypted)
            decrypted = Base64Encrption.decrypt(encrypted)
            self.assertEqual(origin, decrypted)
            
    def test_des(self):
        for test_str in TestEncryption.test_contents:
            origin = test_str.encode()
            encrypted = DES.encrypt(origin)
            # print(encrypted)
            decrypted = DES.decrypt(encrypted)
            self.assertEqual(origin, decrypted)
            
    def test_rsa(self):
        for test_str in TestEncryption.test_contents:
            if len(test_str) < 117:
                origin = test_str.encode()
                encrypted = RSA.encrypt(origin, RSA.pub)
                # print(encrypted)
                decrypted = RSA.decrypt(encrypted, RSA.pri)
                self.assertEqual(origin, decrypted)
        
        



from argparse import ArgumentParser

from .tools.utils import gethostaddr, RandomGen
from .server import start_server
from .client import start_client

def main():
    argparse = ArgumentParser(prog="python -m outlier", description="This is a chat room for your mates", add_help=False)
    argparse.add_argument("-l", "--log", default="INFO", type=str, choices=["DEBUG", "INFO", "ERROR", "debug", "info", "error"])
    argparse.add_argument("-lh", "--loghandler", default=None, type=str)
    argparse.add_argument("-i", "--ip",     required=False, type=str, default=gethostaddr())
    argparse.add_argument("-p", "--port",   required=False, type=int, default=8809)
    argparse.add_argument("-n", "--name",   required=False, type=str)
    argparse.add_argument("-s", "--server", action="store_true")
    
    
    clientargparse = ArgumentParser(prog="python -m outlier", parents=[argparse], conflict_handler="resolve", usage="python3 -m outlier")
    clientargparse.add_argument("-i", "--ip",     required=True, type=str)
    clientargparse.add_argument("-n", "--name",   required=True, type=str)
    clientargparse.add_argument("-k", "--key",    required=False, type=str, default=RandomGen.getrandomvalue()[:6])
    
    kwargs = vars(argparse.parse_args())
    if kwargs.get('server'):
        start_server(**kwargs)
    else:
        kwargs = vars(clientargparse.parse_args())
        start_client(**kwargs)
    

if __name__ == '__main__':
    main()


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-14 18:19:09
FilePath: /outlier/src/outlier/encryption/protocol.py

'''

import base64
import hashlib
import json
from datetime import datetime

import rsa
from pyDes import CBC, PAD_PKCS5, des

KEY = "asdqwezxc"

    

class Encryption:
    @classmethod
    def encrypt(cls, data:bytes, key=KEY) -> bytes:
        hashcode = int(hashlib.md5(key.encode()).hexdigest(), 16) % 256
        out = bytearray()
        for b in data:
            b ^= hashcode
            hashcode = b
            out.append(b)
        return bytes(out)  
    
    @classmethod
    def decrypt(cls, data:bytes, key=KEY) -> bytes:
        hashcode = int(hashlib.md5(key.encode()).hexdigest(), 16) % 256
        out = bytearray()
        for b in data:
            pre_b = b
            b ^= hashcode
            hashcode = pre_b
            out.append(b)
        return bytes(out)  
    

class Base64Encrption(Encryption):
    @classmethod
    def encrypt(cls, content: bytes, key=None) -> bytes:
        bytesflow = base64.b64encode(content)
        return bytesflow

    @classmethod
    def decrypt(cls, flow: bytes, key=None) -> bytes:
        origin = base64.b64decode(flow)
        return origin
    
    
class DES(Encryption):
    SALT = b"outlier_bypassed_you"
    
    
    @classmethod
    def encrypt(cls, content: bytes, key = b"") -> bytes:
        key = (key + DES.SALT)[:8]
        k = des(key, CBC, key, pad=None, padmode=PAD_PKCS5)
        bytesflow = k.encrypt(content, padmode=PAD_PKCS5)
        bytesflow = Base64Encrption.encrypt(bytesflow)
        return bytesflow

    @classmethod
    def decrypt(cls, flow: bytes, key = b"") -> bytes:
        key = (key + DES.SALT)[:8]
        flow = Base64Encrption.decrypt(flow)
        k = des(key, CBC, key, pad=None, padmode=PAD_PKCS5)
        flow = k.decrypt(flow, padmode=PAD_PKCS5)
        return flow

    
class RSA(Encryption):
    pub, pri = rsa.newkeys(1024)
    @classmethod
    def load_public_key(cls, pub: bytes) -> rsa.PublicKey:
        return rsa.PublicKey.load_pkcs1(pub)
    
    @classmethod
    def generate_keys(cls, n=1024):
        RSA.pub, RSA.pri = rsa.newkeys(n)
    
    @classmethod
    def encrypt(cls, content: bytes, pub):
        content = Base64Encrption.encrypt(content)
        return rsa.encrypt(content, pub)
        
    @classmethod
    def decrypt(cls, flow: bytes, pri):
        flow = rsa.decrypt(flow, pri)
        return Base64Encrption.decrypt(flow)
    
    
    
class Package:
    """Package for communication, encapsuling `timestamp`, `cmd` and parameters,
    better be initialized with `Package.buildpackage`
    """
    
    CMD     = "cmd"
    TIME    = "timestamp"
    
    def __init__(self, **data) -> None:
        self.data = data if data is not None else dict()

    def add_field(self, field: str, value) -> 'Package':
        self.data[field] = value
        return self
    
    def add_field_if(self, condition: bool, field: str, value) -> 'Package':
        if condition:
            self.data[field] = value
        return self
    
    def add_cmd(self, cmd: str) -> 'Package':
        self.data[Package.CMD] = cmd
        return self
    
    def encrypt(self, cls = Base64Encrption, key = None) -> bytes:
        try:
            data = json.dumps(self.data).encode()
            return  cls.encrypt(data, key)
        except (json.JSONDecodeError, UnicodeEncodeError) as e:
            return b""
        
    @staticmethod
    def decrypt(byteflow: bytes, cls = Base64Encrption, key = None) -> 'Package':
        try:
            data = cls.decrypt(byteflow, key).decode()
            data = json.loads(data)
            return Package(**data)
        except (json.JSONDecodeError, UnicodeDecodeError) as e:
            return Package()
    
    @staticmethod
    def buildpackage() -> 'Package':
        return Package().add_field(Package.TIME,  datetime.timestamp(datetime.now()))
    
    
    def __repr__(self) -> str:
        return f"{self.data}"  
    
    def __getitem__(a, b):
        return a.data.get(b, None)
    
    def __setitem__(a, b, c):
        a.data[b] = c
            
    def get_field(self, item, default=None):
        ret = self.data.get(item)
        return ret if ret is not None else default
    
        
    


'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-14 19:00:55
FilePath: /outlier/src/outlier/encryption/sessionservice.py

This module is to encrypt and decrypt tcp byteflow and provide better interfaces for business layer
'''
import logging
import time
from dataclasses import dataclass
from typing import Any, Dict

from rsa.pkcs1 import DecryptionError

from ..tools.decorators import onexit, singleton
from ..tools.events import Ops
from ..tools.threadpool import ThreadPool
from ..tools.utils import RandomGen
from ..transmission.tcpservice import Connection, TcpService
from .protocol import DES, RSA, Base64Encrption, Package


@dataclass
class Session:
    """Encapsuled connection and organize it with group
    """
    conn:   Connection  = None
    group:  int         = -1
    secret: str         = None
    state:  int         = 0
    retry:  int         = 0

class BuildChannelException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)
    
class SessionService:
    def __init__(self, service: TcpService, **kwargs) -> None:
        self.tsservice: TcpService = service
        self.tsservice.set_upper_rchandle(self.rchandle)
        self.upper_rchandle = lambda *args: ...
        self.kwargs = kwargs
        
    def set_upper_rchandle(self, upper_rchandle):
        self.upper_rchandle = upper_rchandle
        
    def send(self, package: Package, session: Session = None, cls = Base64Encrption, key = None):
        """General send method, maybe should keep all te parameters here.
        Args:
            package (Package): _description_
            session (Session, optional): _description_. Defaults to None.
            cls (_type_, optional): _description_. Defaults to Base64Encrption.
            key (_type_, optional): _description_. Defaults to None.
        """
        try:
            byteflow = self._convert_package_to_byteflow(package, cls, key)
            self.tsservice.send(byteflow, session.conn)
        except OSError as e:
            logging.warn(f"[Sess layer]\tsend failed {session.conn.addr}")
            
            
    def send_group(self, bcpackage: Package, *sessions: Session):
        pass
        
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        pass
        
    def _convert_byteflow_to_package(self, byteflow: bytes, cls = Base64Encrption, key = None) -> Package:
        return Package.decrypt(byteflow, cls, key) if byteflow else Package.buildpackage()
        
    def _convert_package_to_byteflow(self, package: Package, cls = Base64Encrption, key = None) -> bytes:
        return package.encrypt(cls, key) if package else Package().encrypt(cls, key)
    
    def start(self):
        self.tsservice.startloop()
    
    @onexit
    def close(self, *args):
        logging.debug(f"[Sess layer]\tclose with {args}")
        self.tsservice.close()
        
    
@singleton    
class ServerSessService(SessionService):
    def __init__(self, service: TcpService, **kwargs) -> None:
        super().__init__(service, **kwargs)
        self.sesss: Dict[Any, Session] = dict()
        ThreadPool().put_task(self._switch_rsa)
          
    def _switch_rsa(self):
        import time
        while True:
            RSA.generate_keys()
            time.sleep(60)
    
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        session = self.sesss.get(conn.addr)
        if ops == Ops.Add:
            session = Session(conn)
            self.sesss[conn.addr] = session
        elif ops == Ops.Rmv:
            self.sesss.pop(conn.addr)
        elif ops == Ops.Rcv:
            ...
        # make a secret channel
        # 1. give public key, skip biz handler 
        # 2. confirm secret skip biz handler
        # 3. secret established use biz handler
        package = self._channel_state_convert(session, byteflow, ops)
        logging.debug(f"[Sess layer]\trecall {ops}, {conn.addr}")
        self.upper_rchandle(ops, session, package, *args)
    
    def send(self, package: Package, session: Session = None):
        """to send package to the certain session connecting
        Args:
            package (Package): the sending package
            session (Session, optional): the session where to
        """
        if session.secret:
            super().send(package, session, DES, session.secret)
        else:
            super().send(package, session)
            
                    
    def send_group(self, bcpackage: Package, *sessions: Session):
        for session in sessions:
            if session in self.sesss.values():
                self.send(bcpackage, session)
                
    
    def _channel_state_convert(self, session: Session, byteflow: bytes, ops: Ops) -> Package:
        """server session start conversion
        
        0: init state
        1: waiting secret
        2: waiting ack
        5: check

        if any of 0->1, 1->2, 2->5 fails, back to 0

        Args:
            session (Session): _description_
            byteflow (bytes): _description_
            ops (Ops): _description_
        """
        if ops == Ops.Add:
            session.state = 0
        try:
            if session.state == 0:
                self.send(Package.buildpackage().add_field("session_pub", RSA.pub.save_pkcs1().decode()), session)
                session.state = 1
                session.retry += 1
                if session.retry >=3:
                    self.tsservice._rchandle(Ops.Rmv, session.conn, session.conn.sock.fileno())
                return self._convert_byteflow_to_package(byteflow)
            
            elif session.state == 1:
                # reveive secret
                # send secret
                # waiting for confirm
                package = self._convert_byteflow_to_package(byteflow)
                if package.get_field('session_secret') is not None:
                    secret = package.get_field("session_secret")
                    secret = bytes(bytearray(map(int, secret.split('_'))))
                    secret = RSA.decrypt(secret, RSA.pri)
                    session.secret = secret
                    self.send(Package.buildpackage().add_field("session_confirm", True), session)
                    session.state = 2
                    return package
                else:
                    raise BuildChannelException("no `secret` found")
            elif session.state == 2:
                # receive confirm 
                package = self._convert_byteflow_to_package(byteflow, DES, session.secret)
                if package.get_field('session_ack') is True:
                    session.state = 5
                    return package
                else:
                    raise BuildChannelException("no `ack` found")
            elif session.state == 5:
                return self._convert_byteflow_to_package(byteflow, DES, session.secret)
        except (BuildChannelException, DecryptionError, ValueError) as e:
            from traceback import print_exc
            logging.debug(f"[Sess layer]\tbuild channel failed at {print_exc()} retrying")
            logging.warn(f"[Sess layer]\tbuild channel failed at {e}")
            session.state = 0
            session.secret = None
            self.send(Package.buildpackage().add_field("session_state", -1), session)
        return Package.buildpackage()

                
@singleton
class ConnectSessService(SessionService):
    def __init__(self, service: TcpService, **kwargs) -> None:
        super().__init__(service, **kwargs)
        self.session    = Session(service.conn)
        self.session.state = 1
        self.key: bytes = kwargs.get("key", RandomGen.getrandomvalue()[:6]).encode()
        self.ready      = False
        
    def start(self):
        super().start()
    
        
    def send(self, package: Package, session: Session = None):
        session = session if session is not None else self.session
        if self.session.secret is not None:
            super().send(package, session, DES, self.session.secret)
        else:
            super().send(package, session)
        
    
    def rchandle(self, ops: Ops, conn: Connection, byteflow: bytes = None, *args):
        # 1. parse public key
        # 2. parse confirming secrete key and set secrete key
        # 3. user biz communication
        if byteflow is not None:
            package = self._channel_state_convert(self.session, byteflow, ops)
            self.upper_rchandle(ops, self.session, package, *args)
        
        
    def _channel_state_convert(self, session: Session, byteflow: bytes, ops: Ops) -> Package:
        """client session start conversion
        
        0: backup state, ask for pubkey
        1: waiting for pubkey
        2: waiting for confirm
        5: check

        if any of 0->1, 1->2, 2->5 fails, back to 0

        Args:
            session (Session): _description_
            byteflow (bytes): _description_
            ops (Ops): _description_

        Raises:
            BuildChannelException: _description_
            BuildChannelException: _description_

        Returns:
            Package: _description_
        """
        try:
            package = self._convert_byteflow_to_package(byteflow)
            if package.get_field('session_state') == -1:
                raise BuildChannelException("server confirming failded")
                
            if session.state == 1:
                # reveive secret
                # send secret
                # waiting for confirm
                package = self._convert_byteflow_to_package(byteflow)
                if package.get_field('session_pub') is not None:
                    pub = package.get_field('session_pub')
                    print(f"Get public key {pub}\n")
                    pub = RSA.load_public_key(pub.encode())
                    secret = RSA.encrypt(self.key, pub)
                    print(f"Sending your secret key {self.key} on {secret}\n")
                    self.send(Package.buildpackage().add_field("session_secret", "_".join(str(a) for a in secret)))
                    self.session.secret = self.key
                    session.state = 2
                    return package
                else:
                    raise BuildChannelException("no `session_pub` found")
            elif session.state == 2:
                # receive confirm 
                package = self._convert_byteflow_to_package(byteflow, DES, session.secret)
                if package.get_field('session_confirm'):
                    self.send(Package.buildpackage().add_field("session_ack", True))
                    print(f"Server confirmed your key\n")
                    session.state = 5
                    self.ready = True
                    time.sleep(1)
                    print(f"Channel built\n")
                else:
                    raise BuildChannelException("no `session_confirm` found")
            elif session.state == 5:
                return self._convert_byteflow_to_package(byteflow, DES, session.secret)
        except (Exception, ValueError) as e:
            from traceback import print_exc
            logging.debug(f"[Sess layer]\tbuild channel failed at {print_exc()} retrying")
            session.state = 1
            session.secret = None
            session.retry += 1
            print("--- retry ---")
            if session.retry >= 5:
                self.close()
            else:
                self.send(Package.buildpackage(), session)
        return Package.buildpackage()
    


import os
import signal
import sys
from typing import Tuple, List
from wcwidth import wcswidth
from multiprocessing import Process


class Terminal:
    def __init__(self, log_file_name = None) -> None:
        self.buffer: List[str] = []
        self.history = []
        self.his_idx = 0
        self.cursor_idx = 0
        self.active = True
        self.f = open(log_file_name, 'w') if log_file_name else None
        self.fd = sys.stdout.fileno()

    def _readkey(self) -> Tuple[bytes, bytes]:
        ...
        
    def _up(self):
        if 0 <= self.his_idx - 1 < len(self.history):
            self.his_idx -= 1
            self.buffer.clear()
            self.buffer.extend(list(self.history[self.his_idx]))
            self.rewrite_line(f'\r{self.prompt}' +''.join(self.buffer))
            self.cursor_idx = len(self.buffer)
            sys.stdout.flush()
        
    def _down(self):
        self.buffer.clear()
        if 0 < self.his_idx + 1 < len(self.history):
            self.his_idx += 1
            self.buffer.extend(self.history[self.his_idx])
        elif self.his_idx + 1 == len(self.history):
            self.his_idx += 1
            self.buffer.extend([])
        
        self.rewrite_line(f'\r{self.prompt}' + ''.join(self.buffer))
        self.cursor_idx = len(self.buffer)
        sys.stdout.flush()
            
    def _get_len(self, a: str):
        return wcswidth(a)
        
    def _left(self):
        if self.cursor_idx > 0:
            left_len = self._get_len(self.buffer[self.cursor_idx -1])
            self.cursor_idx -= 1
            for _ in range(left_len):
                os.write(self.fd, b'\x1b[D')
        
    def _right(self):
        if self.cursor_idx < len(self.buffer):
            right_len = self._get_len(self.buffer[self.cursor_idx])
            self.cursor_idx += 1
            for _ in range(right_len):
                os.write(self.fd, b'\x1b[C')
                
    def _reset_cursor_after_full_output(self):
        shift_len = self._get_len("".join(self.buffer[self.cursor_idx:]))
        self._write_spec_char(b'\x1b[D', shift_len)
        # self.cursor_idx = 0
        
    def _home(self):
        shift_len = self._get_len("".join(self.buffer[:self.cursor_idx]))
        self._write_spec_char(b'\x1b[D', shift_len)
        self.cursor_idx = 0
        
    def _end(self):
        shift_len = self._get_len("".join(self.buffer[self.cursor_idx:]))
        self._write_spec_char(b'\x1b[C', shift_len)
        self.cursor_idx = len(self.buffer)
        
    def _insert(self, char: bytes) -> None:
        self.buffer.insert(self.cursor_idx, char.decode())
        self.cursor_idx += 1
        os.write(self.fd, char)
        os.write(self.fd, "".join(self.buffer[self.cursor_idx:]).encode())
        shift_len = self._get_len("".join(self.buffer[self.cursor_idx:]))
        self._write_spec_char(b'\x1b[D', shift_len)
        
    def _delete(self) -> None:
        if 0 <= self.cursor_idx < len(self.buffer):
            self.buffer.pop(self.cursor_idx)
            self._write_spec_char(b'\x1b[K')
            shift_len = self._get_len("".join(self.buffer[self.cursor_idx:]))
            os.write(self.fd, "".join(self.buffer[self.cursor_idx:]).encode())
            self._write_spec_char(b'\x1b[D', shift_len)
        
    def _delete_former(self) -> None:
        if 0 <= self.cursor_idx - 1 < len(self.buffer):
            shift_len = self._get_len(self.buffer[self.cursor_idx - 1])
            self.buffer.pop(self.cursor_idx - 1)
            self._write_spec_char(b'\x1b[D', shift_len)
            self._write_spec_char(b'\x1b[K')
            shift_len = self._get_len("".join(self.buffer[self.cursor_idx - 1:]))
            os.write(self.fd, "".join(self.buffer[self.cursor_idx - 1:]).encode())
            self._write_spec_char(b'\x1b[D', shift_len)
            self.cursor_idx -= 1
        
    def _write_spec_char(self, char: bytes, times: int = 1) -> None:
        if times > 0:
            for _ in range(times):
                os.write(self.fd, char)
        
    def _write_to(self, msg: str):
        if self.f: os.write(self.f.fileno(), f"{msg}\n".encode())
        
    def _write_buffer_details(self):
        self._write_to(f"{self.buffer}, {self.buffer}, {self.cursor_idx}")
        
    def _parse_bytes(self, b: bytes) -> str:
        return "default"
    
    def _input_confirm(self) -> str:
        self.cursor_idx = 0
        outputstr = "".join(self.buffer)
        # self.rewrite_line(f"\r{outputstr}")
        self.buffer.clear()
        sys.stdout.write('\r\n')
        if len(outputstr):
            self.history.append(outputstr)
            self.his_idx = len(self.history)
        sys.stdout.flush()
        self._write_to(f"Check history {self.history}")
        return outputstr
    
    
    def input(self, prompt = ">>> "):
        self.rewrite_line('\r')
        self.rewrite_line(prompt)
        self.prompt = prompt
        while self.active:
            a, b = self._readkey()
            instru = self._parse_bytes(b)
            self._write_to(f"check inputting msg {a} {b} {instru}\n")

            if instru == "backspace":
                self._delete_former()
            elif instru == "delete":
                self._delete()
            elif instru == "home":
                self._home()
            elif instru == "end":
                self._end()
            elif instru == "right":
                self._right()
            elif instru == "left":
                self._left()
            elif instru == "up":
                self._up()
            elif instru == "down":
                self._down()
            elif instru == "enter":
                return self._input_confirm()
            elif instru == "interrupt":
                os.kill(os.getpid(), signal.SIGINT)
            elif instru == "ignore":
                ...
            elif instru == 'default':    
                self._insert(a)
                
            self._write_buffer_details()        
            
    def output(self, output: str) -> None:
        output = str(output)
        output = output.replace('\n', '\r\n')
        sys.stdout.write('\r')
        self.rewrite_line(output)
        sys.stdout.write('\r\n')
        
        self.rewrite_line(self.prompt)
        sys.stdout.flush()
        
        if len(self.buffer):
            sys.stdout.write("".join(self.buffer))
            sys.stdout.flush()
            self._reset_cursor_after_full_output()
            pre_cursor_idx = self.cursor_idx
            self._home()
            for _ in range(pre_cursor_idx):
                self._right()
        sys.stdout.flush()
        self._write_buffer_details()
        
        
        
    def rewrite_line(self, line: str) -> None:
        os.write(self.fd, b'\x1b[2K')
        os.write(self.fd, line.encode())
        
        
    def rewrite_lines(self, *lines: str) -> None:
        lines_all = []
        for l in lines:
            lines_all.extend(l.split("\n"))
        for l in lines_all:
            self.rewrite_line(f"\r{l}\n")
            
    def refresh_lines(self, *lines: str) -> None:
        self._write_spec_char(b'\x1b[2J')
        self.rewrite_lines(*lines)
        self.output("")
        
    def rewrite_rightside(self, line: str) -> None:
        ...

    def __del__(self):
        self.close()

    def close(self):
        self.active = False
        if self.f: self.f.close()


if sys.platform != 'win32':
    import termios
    import tty
    
    # @singleton
    class PosixTerminal(Terminal):
        def __init__(self, log_file_name=None) -> None:
            super().__init__(log_file_name)
            self.valid = False
            try:
                self.settings = termios.tcgetattr(self.fd)
                self.valid = True
            except:
                self.settings = None
                self.valid = False
    
        def _readkey(self) -> Tuple[bytes, bytes]:
            def _readchar() -> bytes:
                if self.valid:
                    try:
                        tty.setraw(self.fd)
                        ch = sys.stdin.read(1)
                    except:
                        print("tty ends")
                    finally:
                        termios.tcsetattr(self.fd, termios.TCSADRAIN, self.settings)
                    return ch.encode()
                else:
                    ch = sys.stdin.read(1)
                    return ch.encode()

            c1 = _readchar()
            if c1 != b'\x1b':
                return c1, c1
            c2 = _readchar()
            if c2 != b'\x5b':
                return c2, c1 + c2
            c3 = _readchar()
            if c3 != b'6' and c3 != b'5' and c3 != b'3':
                return c3, c1 + c2 + c3
            c4 = _readchar()
            return     c4, c1 + c2 + c3 + c4
        
        def _parse_bytes(self, b: bytes) -> str:
            if b == b'\x7f':        return 'backspace'
            if b == b'\r':          return 'enter'
            if b == b'\x1b[3~':     return 'delete'
            if b == b'\x1b[D':      return 'left'
            if b == b'\x1b[C':      return 'right'
            if b == b'\x1b[A':      return 'up'
            if b == b'\x1b[B':      return 'down'
            if b == b'\x1b[H':      return 'home'
            if b == b'\x1b[F':      return 'end'
            if b == b'\x03':        return "interrupt"
            if b[:2] == b'\x1b[':   return 'ignore'
            return super()._parse_bytes(b)
        
        def close(self):
            super().close()
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.settings)
            
    terminal = PosixTerminal()
            
else:
    import msvcrt
    class WinTerminal(Terminal):
        def __init__(self, log_file_name=None) -> None:
            super().__init__(log_file_name)
            self.valid = True
            
        def _readkey(self) -> Tuple[bytes, bytes]:
            def _readchar() -> bytes:
                return msvcrt.getwch().encode()

            c1 = _readchar()
            if c1 != b'\xc3\xa0':
                return c1, c1
            c2 = _readchar()
            return c2, c1 + c2
        
        def _parse_bytes(self, b: bytes) -> str:
            if b == b'\x08':            return 'backspace'
            if b == b'\r':              return 'enter'
            if b == b'\xc3\xa0S':       return 'delete'
            if b == b'\xc3\xa0K':       return 'left'
            if b == b'\xc3\xa0M':       return 'right'
            if b == b'\xc3\xa0H':       return 'up'
            if b == b'\xc3\xa0P':       return 'down'
            if b == b'\xc3\xa0G':       return 'home'
            if b == b'\xc3\xa0O':       return 'end'
            if b == b'\x03':            return "interrupt"
            if b[:-1] == b'\xc3\xa0':   return 'ignore'
            return super()._parse_bytes(b)
        
    terminal = WinTerminal()
    
    
# import asyncio
# import random
# import time 

# async def request(req: str) -> str:
#     await asyncio.sleep(1)
#     # print(f"resp from async {req}")
#     return f"resp from async {req}"

# def randommesssage() -> None:
#     while True:
#         time.sleep(random.random()*5)
#         terminal.output(f"random message {random.random()}")
        
    
# async def client():
#     # asyncio.to_thread(randommesssage)
#     while True:
#         c = terminal.input("$ ")
#         if len(c):
#             r = await request(c)
#             terminal.output(r)

# from threading import Thread

# a = Thread(target=randommesssage)
# a.daemon = True
# a.start()

# asyncio.get_event_loop().run_until_complete(client())



import sys
import time
from argparse import ArgumentParser
from typing import List

from .biz.beans import ChatMessage
from .biz.bizservice import BizRequest, ClientBizService, State, bizclnt
from .encryption.sessionservice import ConnectSessService, SessionService
from .server import Server
from .tools.chatterminal import terminal
from .tools.utils import RandomGen, initlogger
from .transmission.tcpservice import TcpConnectService

if terminal.valid:
    print = terminal.output
    input = terminal.input

class Client(ClientBizService):
    def __init__(self, sessservice: SessionService, **kwargs) -> None:
        super().__init__(sessservice, **kwargs)
        self.buffer: List[ChatMessage] = list()
        self.name = kwargs.get('name')
        self.active = True
        
    def show_msg(self, package):
        print(package.get_field('param'))
        
    def show_chat(self, package):
        cm = package.get_field('param')
        cm = ChatMessage.parse(cm)
        self.buffer.append(cm)
        terminal.refresh_lines(*[a.format() for a in self.buffer])
        # for a in self.buffer:
        #     print(a.format())
    
    def connected(self, package):
        self.atstate = State.Hall
        self.show_msg(package)
        
    def process_input(self, inputs: str):
        super().process_input(inputs)
    
    @bizclnt(state=State.IDLE, bindto=Server.connectuser, recall=connected)
    def connect(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest(param=self.name)
    
    
    @bizclnt(state=State.Room, invokeptn="^(?!\$).+", bindto=Server.chat, recall=show_chat)
    def chat(self, inputs=None, *args, **kwargs) -> BizRequest:
        from datetime import datetime
        cm = ChatMessage(self.name, datetime.now().timestamp(), inputs)
        return BizRequest(param=cm.getattrs())
    
    
    
    @bizclnt(state=State.Hall, invoke="$info", bindto=Server.infohall, recall=show_msg, 
             desc="To glance at the details of the server and client configuartions")
    def info(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    @bizclnt(state=State.Hall, invoke="$exit", 
             desc="To disconnect with the server")
    def quit(self, inputs=None, *args, **kwargs):
        self.close()
        return BizRequest()
        
    
    @bizclnt(state=State.Hall, invoke="$help")
    def help(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    def entered(self, package):
        self.atstate = State.Room
        self.show_msg(package)
    
    @bizclnt(state=State.Hall, invoke="$room", bindto=Server.enterroom, recall=entered,
            descparams="[room name]",
            desc="To enter one of the room or create one by specified name or your username")
    def enter(self, inputs=None, *args, **kwargs) -> BizRequest:
        if len(args) == 1:
            return BizRequest(param=args[0])
        else:
            raise Exception("Parameter Error")
    
    
    
    def left(self, package):
        self.atstate = State.Hall
        self.show_msg(package)

    @bizclnt(state=State.Room, invoke="$leave", bindto=Server.leftroom, recall=left)
    def leave(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    
    @bizclnt(state=State.Room, invoke="$info", bindto=Server.roominfo, recall=show_msg,
            desc="To glance at the details of the room")
    def roominfo(self, inputs=None, *args, **kwargs) -> BizRequest:
        return BizRequest()
    
    
    @bizclnt(state=State.Room, invoke="$clear",
            desc="To clear the terminal")
    def clear(self, inputs=None, *args, **kwargs) -> BizRequest:
        import subprocess
        subprocess.call('reset')
        return BizRequest()
            
        
    def start(self):
        def waituntilready():
            retrials = 0
            while self.active:
                retrials += 1
                time.sleep(3)
                if self.sessservice.ready and self.sessservice.tsservice.loop:
                    break
                print(f"Waiting for channel ready for {retrials * 5} / {5 * 5} secs")
                if retrials > 4:
                    raise Exception("channel build failed")
            retrials = 0
            
        try:
            super().start()
            waituntilready()
            self.connect()
            while self.active:
                a = input()
                self.process_input(a)
        except Exception as e:
            print(e)
            self.close()
            
    def close(self, *args):
        super().close(*args)
        self.active = False
        
def start_client(**kwargs):
    initlogger(kwargs.get('log').upper(), filehandlename=kwargs.get('loghandler'))
    ts = TcpConnectService(False, **kwargs)
    ss = ConnectSessService(ts, **kwargs)
    bs = Client(ss, **kwargs)
    bs.start()

def main():
    argparse = ArgumentParser(prog="Chat room", description="This is a chat room for your mates")
    argparse.add_argument("-l", "--log", default="error", type=str, choices=["DEBUG", "INFO", "ERROR", "debug", "info", "error"])
    argparse.add_argument("-n", "--name", required=True, type=str)
    argparse.add_argument("-i", "--ip",   required=True, type=str)
    argparse.add_argument("-p", "--port", required=False, type=int, default=8809)
    argparse.add_argument("-k", "--key",  required=False, type=str, default=RandomGen.getrandomvalue()[:6])
    kwargs = vars(argparse.parse_args())
    start_client(**kwargs)

    
if __name__ == '__main__':
    main()



'''
Date: 2023-03-08 23:10:22
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-13 19:34:21
FilePath: /outlier/test/test_threadpool.py
'''
import sys
import time
import unittest
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent / "src" ))

from outlier.tools.threadpool import ThreadPool

class testThreadPool(unittest.TestCase):
    def test_norm(self):
        starttime = time.time()
        t = ThreadPool(10, False)
        t.put_task(ThreadPool.template_loop, args=(1,))
        t.put_task(ThreadPool.template_loop, args=(5,))
        t.put_task(ThreadPool.template_loop, args=(2,))
        t.put_task(ThreadPool.template_loop, args=(4,))
        t.put_task(ThreadPool.template_loop, args=(3,))
        t.queue.join()
        endtime = time.time()
        self.assertAlmostEqual(endtime - starttime, 5, delta=0.9)
        
    def test_thread_dead(self):
        starttime = time.time()
        t = ThreadPool(2, False)
        t.put_task(lambda x: 1/ x, args=(0,))
        t.put_task(ThreadPool.template_loop, args=(5,))
        time.sleep(2)
        t.put_task(ThreadPool.template_loop, args=(2,))
        t.queue.join()
        endtime = time.time()
        self.assertAlmostEqual(endtime - starttime, 5, delta=0.9)



'''
Date: 2023-03-13 11:04:33
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-14 17:57:14
FilePath: /outlier/src/outlier/__init__.py
'''

__all__ = ['server', 'client']

from . import server
from . import client
